if string then
	function string.contains(a, b) return a:find(b, 1, true) end
end

function encode_thing(x)
	if type(x) == "number" then return tostring(x)
	elseif type(x) == "string" then return string.format("%q", x):gsub("\\\n", "\n")
	elseif type(x) == "boolean" then
		if x == true then
			return "True"
		else
			return "False"
		end
	elseif type(x) == "table" then
		return table_to_str(x)
	end
end

function decode_string(x) -- TODO-future: can this be done in a better way?
	f = loadstring("return " .. x)
	if f then
		return f()
	else
		return nil
	end
end

function decode_thing(x)
	local num = tonumber(x)
	if num then return num end
	if x == "True" then return true end
	if x == "False" then return false end
--~ 	if x == nil then -- TODO: Hack, it's a bug when this happens
--~ 		print("Error decoding string, x = nil")
--~ 		return "nil"
--~ 	end
	local str = x:match([[^"(.*)"$]]) -- TODO-future: redo these in some other way?
	if str then
		s = decode_string(x)
		if s then return s end
	end
	local tblstr = x:match([[^(%[.+%])$]]) -- TODO-future: redo these in some other way?
	if tblstr then return str_to_table(tblstr) end
	local eff1, eff2, eff3 = x:match([[^%("(.-)", *([0-9]-), *"(.-)"%)$]])
	if eff1 and eff2 and eff3 then
		return { eff1, eff2, eff3 }
	end
	print("error: can't decode table value", x)
end

function table_to_str(tbl)
	local newtbl = {}
	local keys = {}
	for from, to in pairs(tbl) do
		table.insert(keys, from)
	end
	table.sort(keys)
	for _, from in pairs(keys) do
		local to = tbl[from]
--~ 		print("tbl["..tostring(from).."] -> "..tostring(to))
		table.insert(newtbl, string.format("(%s, %s)", encode_thing(from), encode_thing(to)))
	end
	return "["..table.concat(newtbl, ", ").."]"
end

function str_to_table(str) -- TODO: redo properly!
	local tbl = {}
-- 	print("str_to_table("..str..")")
	for x in str:gmatch("%b()") do
--~ 		print(x)
		local a, b = x:match([[^%(("[^"]*,[^"]*"), *([0-9]+)%)$]]) -- ugly hack for e.g. Go Get 'Em, Tiger!
		if not b then
			a, b = x:match([[^%((.-), *(.+)%)$]]) -- TODO-future: redo regex?
		end
-- 		print("decoding", a, "->", b)
		tbl[decode_thing(a)] = decode_thing(b)
	end
--~ 	print("returning", table_to_str(tbl))
	return tbl
end

function parse_params_raw(str)
	local loadstr = parse_table_string(str)
	if loadstr == "" then
		return nil
	else
		local tbl = assert(loadstring(loadstr))()
		if table.maxn(tbl) == 0 then
			return nil
		else
			return tbl
		end
	end
end

function parse_params(str)
	local rawtbl = parse_params_raw(str)
	local tbl = {}
	for _, x in pairs(rawtbl or {}) do
		tbl[x.key] = x.value
	end
	return tbl
end

function printstr(tbl)
	if tbl == nil then
		return tostring(tbl)
	elseif type(tbl) == "table" then
		return table_to_str(tbl)
	else
		return tostring(tbl)
	end
end

function get_daily_counter(name)
	local value = get_day_state(name)
	if (value == "") then value = 0 end
	return tonumber(value)
end

function increase_daily_counter(name, amount)
	local increase = (amount or 1)
	set_day_state(name, get_daily_counter(name) + increase)
end

function reset_daily_counter(name)
	set_day_state(name, "")
end

local familiar_counters = {}
local infolines = {}
local values = {}

function print_familiar_counter(compact, normal)
	if text:contains("<!-- charpane compact") then
		table.insert(familiar_counters, compact)
	else
		table.insert(familiar_counters, normal)
	end
end

function print_charpane_infoline(compact, normal)
	if text:contains("<!-- charpane compact") then
		table.insert(infolines, compact)
	else
		table.insert(infolines, normal)
	end
end

function print_charpane_value(value)
	table.insert(values, value)
end

function print_charpane_lines()
	if text:contains("http://images.kingdomofloathing.com/otherimages/inf_small.gif") then return end
	if text:contains("<!-- charpane compact") then -- compact mode
		local lines = {}
		for x,y in pairs(familiar_counters) do table.insert(lines, y) end
		for x,y in pairs(infolines) do table.insert(lines, y) end

		for x,y in pairs(lines) do
			if text:match("</table>(<!%-%- charpane compact familiar text space type%b{} weight%b{} %-%->)") then
				text = text:gsub("(<!%-%- charpane compact familiar text space type%b{} weight%b{} %-%->)", function(one) return y .. one end)
			else
				text = text:gsub("(<!%-%- charpane compact familiar text space type%b{} weight%b{} %-%->)", function(one) return "<br>" .. y .. one end)
			end
		end
		for x,y in pairs(values) do
			local name = y.compactname or y.name or "{nil}"
			local value = y.value or "{nil}"
			local ct = "<td align=right>" .. name .. ":</td><td align=left><b>" .. value .. "</b></td>"
			local style = ""
			if y["color"] then
				style = " style=\"color:" .. y["color"] .. "\""
			end
			if y["link"] then
				ct = [[<td align=right><a target="mainpane" href="]] .. y["link"] .. [["]] .. style .. [[>]] .. name .. [[</a>:</td><td align=left><a target="mainpane" href="]] .. y["link"] .. [["]] .. style .. [[><b>]] .. value .. [[</b></a></td>]]
			end
			if y["tooltip"] then
				ct = "<tr title=\"" .. y["tooltip"] .. "\"" .. style .. ">" .. ct .. "</tr>"
			else
				ct = "<tr" .. style .. ">" .. ct .. "</tr>"
			end
			text = string.gsub(text, "(<!%-%- charpane compact text space %-%->)", function(one) return ct .. one end)
		end
	else -- normal mode
		local lines = {}
		for x,fc in pairs(familiar_counters) do table.insert(lines, tostring(fc)) end
		for x,il in pairs(infolines) do table.insert(lines, tostring(il)) end

		for x,l in pairs(lines) do
			text = text:gsub("(<!%-%- charpane normal familiar text space type%b{} weight%b{} %-%->)", function(one) return "<tr><td colspan=\"2\"><center><font size=\"2\">" .. l .. "</font></center></td></tr>" .. one end)
		end
		for x,y in pairs(values) do
			local name = y.normalname or y.name or "{nil}"
			local value = y.value or "{nil}"
			local nt = [[<font size="2">]] .. name .. ": <b>" .. value .. "</b></font>"
			local style = ""
			if y.color then
				style = [[ style="color:]] .. y.color .. [["]]
			end
			if y.link then
				nt = [[<font size="2"><a target="mainpane" href="]] .. y.link .. [["]] .. style .. [[>]] .. name .. [[: <b>]] .. value .. [[</b></a></font>]]
			end
			if y.tooltip then
				nt = [[<span title="]] .. y.tooltip .. [["]] .. style .. ">" .. nt .. "</span>"
			else
				nt = "<span" .. style .. ">" .. nt .. "</span>"
			end
			text = text:gsub("(<!%-%- charpane normal text space %-%->)", function(one) return nt .. "<br>" .. one end)
		end
	end
end

--~ function get_request_state(name) return get_state("request", name) end
function get_fight_state(name) return get_state("fight", name) end
function get_session_state(name) return get_state("session", name) end
function get_day_state(name) return get_state("day", name) end
function get_ascension_state(name, value) return get_state("ascension", name) end
function get_character_state(name) return get_state("character", name) end

--~ function set_request_state(name, value) set_state("request", name, value) end
function set_fight_state(name, value) set_state("fight", name, value) end
function set_session_state(name, value) set_state("session", name, value) end
function set_day_state(name, value) set_state("day", name, value) end
function set_ascension_state(name, value) set_state("ascension", name, value) end



local function setup_state_table(getf, setf)
	local tbl = {}
	setmetatable(tbl, { __index = function(t, k)
		local function parse_value(v)
			if v == "" then return nil end
			local pref = v:sub(1, 1)
			if pref == "[" then
				return str_to_table(v)
			else
				return v
			end
		end
		local v = getf(k)
-- 		print("getting state[" .. tostring(k) .. "]: " .. tostring(v))
		local p = parse_value(v)
-- 		print("getting state[" .. tostring(k) .. "] => " .. tostring(p))
		return p
	end, __newindex = function(t, k, v)
-- 		print("setting state[" .. tostring(k) .. "]: " .. tostring(v))
		local p = nil
		if v == nil then
			p = ""
		elseif type(v) == "table" then
			p = table_to_str(v)
		elseif type(v) == "number" then
			p = tostring(v)
		elseif type(v) == "string" then
			p = v
		else
			error("Unknown value " .. tostring(v) .. " of type " .. type(v))
		end
-- 		print("setting state[" .. tostring(k) .. "] => " .. tostring(p))
		setf(k, p)
	end})
	return tbl
end

fight = setup_state_table(get_fight_state, set_fight_state)
session = setup_state_table(get_session_state, set_session_state)
day = setup_state_table(get_day_state, set_day_state)
ascension = setup_state_table(get_ascension_state, set_ascension_state)
character = setup_state_table(get_character_state, set_character_state)





after_pageload_cache = {}

function get_cached_function(f)
-- 	print("get_cached_function", f)
	local v = after_pageload_cache[f]
	if v then
		return v
	else
		v = f()
		after_pageload_cache[f] = v
		return v
	end
end

raw_submit_page = submit_page

function do_submit_page(t, url, params)
	kolproxy_debug_print("> do_submit_page()\n" .. debug.traceback(""))
-- 	print("> do_submit_page()\n" .. debug.traceback(""))
	after_pageload_cache = {}
	local pt, pturl = nil, nil
	if params then
		local tbl = {}
		for a, b in pairs(params) do
			if type(b) == "string" then
				table.insert(tbl, { key = a, value = b })
			elseif type(b) == "number" then
				table.insert(tbl, { key = a, value = tostring(b) })
			else
				error("Unknown submit_page value type: " .. type(b))
			end
		end
		pt, pturl = raw_submit_page(t, url, tbl)
	else
		pt, pturl = raw_submit_page(t, url, nil)
	end
	kolproxy_debug_print("< do_submit_page()")
-- 	print("< do_submit_page()")
	return pt, pturl
end


function get_page(url, params) return do_submit_page("GET", url, params) end

function post_page(url, params) return do_submit_page("POST", url, params) end

raw_make_href = make_href

function do_make_href(url, params)
	if params then
		local tbl = {}
		for a, b in pairs(params) do
			table.insert(tbl, { key = a, value = b })
		end
		return raw_make_href(url, tbl)
	else
		return raw_make_href(url, nil)
	end
end


function get_itemid(name)
	local data = get_item_data_by_name(name)
	if not data or not data.id then
		error("No itemid found for item: " .. name)
	end
	return tonumber(data.id)
end

-- function caching_get_itemid(name) -- TODO: would be better if get_item_data_by_name wasn't slow
-- 	local tbl = get_session_state("cached itemids")
-- 	if (tbl == "") then tbl = {} else tbl = str_to_table(tbl) end
-- 	if not tbl[name] then
-- 		tbl[name] = tonumber(str_to_table(get_item_data_by_name(name)).id)
-- 		set_session_state("cached itemids", table_to_str(tbl))
-- 	end
-- 	num = tbl[name]
-- 	if num == nil then
-- 		error("No itemid found for item: " .. name)
-- 	end
-- 	return num
-- end

-- if set_state then
-- 	get_itemid = caching_get_itemid
-- end







dod_potion_types = {
	"bubbly potion",
	"cloudy potion",
	"dark potion",
	"effervescent potion",
	"fizzy potion",
	"milky potion",
	"murky potion",
	"smoky potion",
	"swirly potion",
}

dod_potion_effects = {
	"acuity",
	"blessing",
	"booze",
	"confusion",
	"detection",
	"healing",
	"sleep",
	"strength",
	"teleportation",
}

function get_dod_potion_status()
--~ 	print("get potion status")
	local tbl = get_ascension_state("zone.dod.potions")
	if (tbl == "") then tbl = {} else tbl = str_to_table(tbl) end
	unknown_potions = {}
	for _, pot in pairs(dod_potion_types) do
		if not tbl[pot] then
			table.insert(unknown_potions, pot)
		end
	end

	found = {}
	for pot, eff in pairs(tbl) do
		found[eff] = pot
	end

	unknown_effects = {}
	for _, eff in pairs(dod_potion_effects) do
		if not found[eff] then
			table.insert(unknown_effects, eff)
		end
	end

	if table.maxn(unknown_effects) == 1 and table.maxn(unknown_potions) == 1 then -- only one missing effect and one missing potion
		tbl[unknown_potions[1]] = unknown_effects[1]
		return tbl, {}
	else
--~ 		print("unknown effects", table.maxn(unknown_effects), table_to_str(unknown_effects))
--~ 		print("unknown potions", table.maxn(unknown_potions), table_to_str(unknown_potions))
		return tbl, unknown_potions
	end
end

function add_itemdrop_counter(name, f)
	add_printer("item drop: " .. name, function ()
		c = count(item_name)
		text = text .. f(c)
	end)
end

function intercept_warning(message, warningid)
	if get_session_state(warningid) ~= "" then return end
	pwd = get_session_state("pwd") -- Inserting pwd, boo!
	text = [[<html><head><script type="text/javascript" src="http://images.kingdomofloathing.com/scripts/jquery-1.3.1.min.js"></script>
<script type="text/javascript">
	function disable_warning(link) {
		var pwd = ']] .. pwd .. [['
		var stateset = 'session'
		var name = ']] .. warningid .. [['
		var value = 'skip'
		$.post('custom-settings', { pwd:pwd, action:'set state', name:name, stateset:stateset, value:value, ajax: 1 }, function (res) {
			link.style.color = 'gray'
			link.innerHTML = '(Warning disabled until you log out.)'
		});
	}
</script></head>]] .. [[
<p>Warning: ]] .. message .. [[</p><p>]] ..
[[<a href="#" onclick="javascript:disable_warning(this); return false;" style="color: green;">Disable this warning.</a>]] ..
[[</p><p>]] .. [[<a href="]]..raw_make_href(path, parse_params_raw(input_params))..[[">Repeat request</a>]] .. [[</p><body></body></html>]]
	return text, "/kolproxy-warning"
end

local zone_checks = {}
local zone_check_added = false
local zone_check_tableinsert = table.insert -- HACK: This is a workaround, to get access to table.insert in the zone check function
function add_zone_check(zoneid, check)
	if not zone_check_added then
		add_interceptor("/adventure.php", function()
			if get_character_state("enable adventure warnings") ~= "yes" then return end
			local zid = tonumber(params.snarfblat)
			for xidx, xf in pairs(zone_checks[zid] or {}) do
				-- NB: Zone warnings are identified by zone and order added. This is a bit tricky, it changes if new ones are added.
				local x = xf()
				if x then
					local a, b = intercept_warning(x, "zonecheck-" .. zid .. "-" .. xidx)
					if a then return a, b end
				end
			end
		end)
		zone_check_added = true
	end
	if not zone_checks[zoneid] then zone_checks[zoneid] = {} end
	zone_check_tableinsert(zone_checks[zoneid], check)
end

function run_file_with_environment(filename, env, f_notfound)
	-- HACK: This is a weird workaround, to get util functions to refer to our environment
	env.add_itemdrop_counter = add_itemdrop_counter
	env.add_zone_check = add_zone_check
	setfenv(add_itemdrop_counter, env)
	setfenv(add_zone_check, env)

	setmetatable(env, { __index = function(t, k)
		if f_notfound then
			f_notfound(t, k, filename)
--~ 			print("f_notfound", t, k, _G[k])
			return _G[k]
		end
	end})
--~ 	setmetatable(env, { __index = _G })
	local f, e = loadfile("scripts/" .. filename)
	if not f then error(e, 2) end
	setfenv(f, env)

	f()
end

--~ function run_file_with_environment(filename, env)
--~ 	for a,b in pairs(env) do
--~ 		_G[a] = b
--~ 	end
--~ 	dofile("scripts/" .. filename)
--~ end

function load_script_files(env)
	local function load_file(category, name)
		if category ~= "chat" then
			local warn = true
			run_file_with_environment(name, env, function (t, k, filename)
				if warn then
					print("Warning: invalid __index for name", k, "in", filename)
--~ 					error("Warning: invalid __index for name '" .. tostring(k) .. "' in '" .. tostring(filename) .. "'")
				end
			end)
			warn = false
		end
	end

	run_file_with_environment("loaders.lua", { load_file = load_file })
	run_file_with_environment("user-loaders.lua", { load_file = load_file })
end


function set_ascension_turn_counter(name, length)
	local tbl = get_ascension_state("turn counters")
	if (tbl == "") then tbl = {} else tbl = str_to_table(tbl) end
	table.insert(tbl, table_to_str{ name = name, turn = tonumber(get_turns_this_run()), length = length })
	set_ascension_state("turn counters", table_to_str(tbl))
end

function get_drunkenness()
	print("get_drunkenness() is deprecated, use drunkenness()")
	return drunkenness()
end

function do_we_have_mall_access()
	print("do_we_have_mall_access() is deprecated, use ascensionstatus()")
	return (ascensionstatus() == "Aftercore")
end

function get_familiar_weight()
	print("get_familiar_weight() is deprecated, use buffedfamiliarweight()")
	return buffedfamiliarweight()
end



function setup_functions()
	local function setup_functions_raw()
		function refresh_status()
			print("\n\nrefresh_status() is no longer needed\n\n" .. debug.traceback("") .. "\n\n")
		end
		function status()
			return get_cached_function(get_status_info)
		end
		function buffslist()
			local tbl = {}
			for _, x in pairs(status().effects) do
				tbl[x[1]] = tonumber(x[2])
			end
			return tbl
		end
		function refresh_inventory()
			print("\n\nrefresh_inventory() is no longer needed\n\n" .. debug.traceback("") .. "\n\n")
		end
		function inventory()
			return get_cached_function(get_inventory_counts)
		end

	--~ function level() return tonumber(status().level) end -- didn't work before because CDM moyered it
		function level()
			if basemainstat() < 4 then
				return 1
			else
				return 1 + math.floor(math.sqrt(basemainstat() - 4))
			end
		end

		function classid() return tonumber(status().class) end

		function playerid() return tonumber(status().playerid) end

		function get_mainstat()
			local classname = classid()
			if classname == 1 or classname == 2 then
				return "Muscle"
			elseif classname == 3 or classname == 4 then
				return "Mysticality"
			elseif classname == 5 or classname == 6 then
				return "Moxie"
			end
		end

		-- WARNING: Buffed values can be out of date unless you load charpane.php. This is a KoL/CDM bug.
		function playername() return status().name end
		function meat() return tonumber(status().meat) end
		function hp() return tonumber(status().hp) end
		function mp() return tonumber(status().mp) end
		function maxhp() return tonumber(status().maxhp) end
		function maxmp() return tonumber(status().maxmp) end
		function turnsthisrun() return tonumber(status().turnsthisrun) end
		function familiarid() return tonumber(status().familiar) end
		function buffedfamiliarweight() return tonumber(status().famlevel) end
		function buff(name) return buffslist()[name] end
		function buffturns(name) return buffslist()[name] or 0 end
		function advs() return tonumber(status().adventures) end
		function buffedmuscle() return tonumber(status().muscle) end
		function buffedmysticality() return tonumber(status().mysticality) end
		function buffedmoxie() return tonumber(status().moxie) end
		function basemuscle() return tonumber(status().basemuscle) end
		function basemysticality() return tonumber(status().basemysticality) end
		function basemoxie() return tonumber(status().basemoxie) end
		function buffedmainstat() return tonumber(status()[string.lower(get_mainstat())]) end
		function basemainstat() return tonumber(status()["base" .. string.lower(get_mainstat())]) end
		function rawmainstat() return tonumber(status()["raw" .. string.lower(get_mainstat())]) end
		function ascensionpath() return status().path end
		function moonsign() return status().sign end
		function moonsign_area()
			local areas = {
				Mongoose = "Degrassi Knoll",
				Wallaby = "Degrassi Knoll",
				Vole = "Degrassi Knoll",
				Platypus = "Little Canadia",
				Opossum = "Little Canadia",
				Marmot = "Little Canadia",
				Wombat = "Gnomish Gnomad Camp",
				Blender = "Gnomish Gnomad Camp",
				Packrat = "Gnomish Gnomad Camp",
			}
			return areas[moonsign()]
		end
		function equipment()
			local eq = {}
			for a, b in pairs(status().equipment) do
				eq[a] = tonumber(b)
			end
			eq.fakehands = nil -- Work around API misfeature - this is not an itemid
			return eq
		end
		function fullness() return tonumber(get_daily_counter("fullness")) end
		function drunkenness() return tonumber(status().drunk) end
		function spleen() return tonumber(get_daily_counter("spleen")) end
		function ascensionstatus()
			if tonumber(status().freedralph) == 1 then
				return "Aftercore"
			elseif tonumber(status().casual) == 1 then
				return "Aftercore"
			elseif tonumber(status().hardcore) == 1 then
				return "Hardcore"
			elseif tonumber(status().roninleft) > 0 then
				return "Softcore"
			else
				return "Aftercore"
			end
		end
		function mcd() return tonumber(status().mcd) end

		function have_inventory(name)
			if inventory()[get_itemid(name)] then
				return true
			else
				return false
			end
		end

		function have_equipped_itemid(id)
			for _, x in pairs(equipment()) do
				if x == id then
					return true
				end
			end
			return false
		end

		function have_itemid(id)
			if inventory()[id] then
				return true
			else
				return have_equipped_itemid(id)
			end
		end

		function have(name)
			return have_itemid(get_itemid(name))
		end

		function have_equipped(name)
			return have_equipped_itemid(get_itemid(name))
		end

		function count_itemid(id)
			return inventory()[id] or 0
		end

		function count(name)
			return count_itemid(get_itemid(name))
		end

		function hottub()
			return get_page("/clan_viplounge.php", { action = "hottub" })
		end

		function plunge_items(a, b)
			local pwd = get_session_state("pwd") -- inserting pwd, boo!
			return post_page("/knoll.php", { action = "combine", pwd = pwd, item1 = tostring(get_itemid(a)), item2 = tostring(get_itemid(b)), quantity = "1", ajax = "1" })
		end

		function meatpaste_items(a, b)
			local pwd = get_session_state("pwd") -- inserting pwd, boo!
			post_page("/craft.php", { pwd = pwd, action = "makepaste", qty = "1", ajax = "1", whichitem = get_itemid("meat paste") })
			return post_page("/craft.php", { mode = "combine", pwd = pwd, action = "craft", a = tostring(get_itemid(a)), b = tostring(get_itemid(b)), qty = "1", ajax = "1" })
		end

		function cook_items(a, b)
			local pwd = get_session_state("pwd") -- inserting pwd, boo!
			return post_page("/craft.php", { mode = "cook", pwd = pwd, action = "craft", a = tostring(get_itemid(a)), b = tostring(get_itemid(b)), qty = "1", ajax = "1" })
		end

		function mix_items(a, b)
			local pwd = get_session_state("pwd") -- inserting pwd, boo!
			return post_page("/craft.php", { mode = "cocktail", pwd = pwd, action = "craft", a = tostring(get_itemid(a)), b = tostring(get_itemid(b)), qty = "1", ajax = "1" })
		end

		function smith_items(a, b)
			local pwd = get_session_state("pwd") -- inserting pwd, boo!
			return post_page("/knoll.php", { action = "smith", pwd = pwd, item1 = tostring(get_itemid(a)), item2 = tostring(get_itemid(b)), quantity = "1", ajax = "1" })
		end

		function buy_item(name, whichstore, amount)
			local pwd = get_session_state("pwd") -- inserting pwd, boo!
			amount = amount or 1
			return get_page("/store.php", { phash = pwd, buying = "1", whichitem = tostring(get_itemid(name)), howmany = tostring(amount), whichstore = whichstore, ajax = "1", action = "buyitem" })
		end

		function equip_itemid(id, slot)
			local pwd = get_session_state("pwd") -- inserting pwd, boo!
			local tbl = { pwd = pwd, action = "equip", whichitem = tostring(id), ajax = "1" }
			if slot == "dualwield" then
				tbl.action = "dualwield"
			elseif slot == "offhand" then
				tbl.action = "dualwield"
				local a, b = get_page("/inv_equip.php", tbl)
				if equipment().offhand ~= id then
					tbl.action = "equip"
					tbl.slot = tostring(slot)
					return get_page("/inv_equip.php", tbl)
				else
					return a, b
				end
			elseif slot == "acc1" then
				tbl.slot = "1"
			elseif slot == "acc2" then
				tbl.slot = "2"
			elseif slot == "acc3" then
				tbl.slot = "3"
			elseif slot then
				tbl.slot = tostring(slot)
			end
			return get_page("/inv_equip.php", tbl)
		end

		function equip_item(name, slot)
			return equip_itemid(get_itemid(name), slot)
		end

		function unequip_slot(name)
			local pwd = get_session_state("pwd") -- inserting pwd, boo!
			return get_page("/inv_equip.php", { pwd = pwd, action = "unequip", type = name, ajax = "1" })
		end

		function set_mcd(amount)
			local pwd = get_session_state("pwd") -- inserting pwd, boo!
			-- TODO: non-muscle signs
			return get_page("/inv_use.php", { pwd = pwd, whichitem = tostring(get_itemid("detuned radio")), ajax = "1", tuneradio = tostring(amount) })
		end

		function set_equipment(tbl)
-- 			print("setting equipment to", table_to_str(tbl))
			local eq = equipment()
			for a, b in pairs(eq) do
-- 				print("checking", b, "vs[", a, "]", tbl[a])
				if b ~= tbl[a] then
					unequip_slot(a)
-- 					print("unequipping slot", a)
				end
			end
			local eq = equipment()
			for a, b in pairs(tbl) do
				if eq[a] ~= b then
-- 					print("equip_itemid", b, a)
					equip_itemid(b, a)
				end
			end
			local eq = equipment()
			for a, b in pairs(eq) do
				if b ~= tbl[a] then
					error("Wearing " .. tostring(a) .. ":" .. tostring(b) .. " after trying to wear " .. tostring(tbl[a]))
				end
			end
			for a, b in pairs(tbl) do
				if eq[a] ~= b then
					error("Wearing " .. tostring(a) .. ":" .. tostring(eq[a]) .. " after trying to wear " .. tostring(b))
				end
			end
		end

		function use_itemid(id, amount)
			local pwd = get_session_state("pwd") -- inserting pwd, boo!
			if amount then
				return get_page("/multiuse.php", { pwd = pwd, whichitem = tostring(id), ajax = "1", quantity = tostring(amount), action = "useitem" })
			else
				return get_page("/inv_use.php", { pwd = pwd, whichitem = tostring(id), ajax = "1" })
			end
		end

		function use_itemid_noajax(id, amount)
			local pwd = get_session_state("pwd") -- inserting pwd, boo!
			if amount then
				return get_page("/multiuse.php", { pwd = pwd, whichitem = tostring(id), quantity = tostring(amount), action = "useitem" })
			else
				return get_page("/inv_use.php", { pwd = pwd, whichitem = tostring(id) })
			end
		end

		function use_item(name, amount)
			return use_itemid(get_itemid(name), amount)
		end

		function use_item_noajax(name, amount)
			return use_itemid_noajax(get_itemid(name), amount)
		end

		function eat_item(name)
			local pwd = get_session_state("pwd") -- inserting pwd, boo!
			return get_page("/inv_eat.php", { pwd = pwd, which = "1", whichitem = tostring(get_itemid(name)), ajax = "1" })
		end

		function drink_itemid(id)
			local pwd = get_session_state("pwd") -- inserting pwd, boo!
			return get_page("/inv_booze.php", { pwd = pwd, whichitem = id, ajax = 1 })
		end

		function drink_item(name)
			drink_itemid(get_itemid(name))
		end

		function pull_itemids(xs)
			local pwd = get_session_state("pwd") -- inserting pwd, boo!
			local tbl = { pwd = pwd, action = "pull" }
			for nr, id in pairs(xs) do
				tbl["howmany" .. tostring(nr)] = "1"
				tbl["whichitem" .. tostring(nr)] = tostring(id)
			end
			-- TODO!: split up more than 11
			return post_page("/storage.php", tbl)
		end

		function freepull_itemid(id)
			local pwd = get_session_state("pwd") -- inserting pwd, boo!
			return post_page("/storage.php", { action = "pull", pwd = pwd, howmany1 = 1, whichitem1 = id })
		end

		function closet_itemid(id)
			local pwd = get_session_state("pwd") -- inserting pwd, boo!
			return get_page("/inventory.php", { action = "closetpush", pwd = pwd, qty = 1, whichitem = id, ajax = 1 })
		end

		function uncloset_itemid(id)
			local pwd = get_session_state("pwd") -- inserting pwd, boo!
			return get_page("/inventory.php", { action = "closetpull", pwd = pwd, qty = 1, whichitem = id, ajax = 1 })
		end
		
		function sell_itemid(id, amount)
			local pwd = get_session_state("pwd") -- inserting pwd, boo!
			return post_page("/sellstuff_ugly.php", { pwd = pwd, action = "sell", mode = 3, quantity = amount, ["item"..tostring(id)] = tostring(id) })
		end

		function cast_buff(skillid, quantity, targetid)
			local pwd = get_session_state("pwd") -- inserting pwd, boo!
			targetid = targetid or playerid()
			assert(targetid and targetid ~= "")
			local tbl = { whichskill = tostring(skillid), ajax = "1", action = "Skillz", pwd = pwd, targetplayer = tostring(targetid) }
			if quantity then tbl.quantity = tostring(quantity) end
			return get_page("/skills.php", tbl)
		end

		function switch_familiarid(id)
			return get_page("/familiar.php", { action = "newfam", ajax = "1", newfam = tostring(id) })
		end

		function handle_adventure_result(pt, url, zoneid, macro, noncombatchoices, specialnoncombatfunction)
			if url:contains("/fight.php") then -- TODO: hmmmm? -- and url:match("ireallymeanit") then
				local advagain = nil
				if pt:contains([[>You win the fight!<!--WINWINWIN--><]]) then
					advagain = true
				elseif pt:contains([[state['fightover'] = true;]]) then
					if pt:match("You lose.") then
						advagain = false
					elseif pt:contains([[<a href="adventure.php?snarfblat=]]..zoneid..[[">Adventure Again]]) then
						advagain = true
					end
				end
				if advagain == nil then
					if macro then
						local pt, url = post_page("/fight.php", { action = "macro", macrotext = macro })
--~ 						print("recurse with macro")
						return handle_adventure_result(pt, url, zoneid, nil, noncombatchoices, specialnoncombatfunction)
					else
						print("fight.php unhandled url", url)
					end
				end
--~ 				print("return1 p u a", pt:len(), url, advagain)
				return pt, url, advagain
			elseif url:contains("/choice.php") then
				local advagain = nil
				local adventure_title = pt:match([[<tr><td style="color: white;" align=center bgcolor=blue><b>([^<]*)</b></td></tr>]])
				if adventure_title == "Results:" then
					if pt:contains([[<a href="adventure.php?snarfblat=]]..zoneid..[[">Adventure Again]]) then
						advagain = true
					end
	--~ 				print("return2 p u a", pt:len(), url, advagain)
					return pt, url, advagain
				end
				local choice_adventure_number = tonumber(pt:match([[<input type=hidden name=whichchoice value=([0-9]+)>]]))
		--~ 		print("choice", adventure_title, choice_adventure_number)
				local pickchoice = nil
				local optname = nil
				if specialnoncombatfunction then
					optname, pickchoice = specialnoncombatfunction(adventure_title, choice_adventure_number)
				else
					optname = noncombatchoices[adventure_title]
				end
				if optname and not pickchoice then
					for nr, title in pt:gmatch([[<input type=hidden name=option value=([0-9])><input class=button type=submit value="([^>]+)">]]) do
			--~ 			print("opt", nr, title)
						if title == optname then
							pickchoice = tonumber(nr)
						end
					end
				end
				if optname and not pickchoice then
					print("Warning: option " .. tostring(optname) .. " not found for " .. tostring(adventure_title) .. ".")
				end
				if pickchoice then
					local pwd = get_session_state("pwd") -- inserting pwd, boo!
					local pt, url = post_page("/choice.php", { pwd = pwd, whichchoice = tostring(choice_adventure_number), option = tostring(pickchoice) })
					print("choice ->", url)
					return handle_adventure_result(pt, url, zoneid, macro, noncombatchoices, specialnoncombatfunction)
				else
					print("choice", adventure_title, choice_adventure_number)
					for nr, title in pt:gmatch([[<input type=hidden name=option value=([0-9])><input class=button type=submit value="([^>]+)">]]) do
						print("opt", nr, title)
					end
	--~ 				print("return3 p u a", pt:len(), url, advagain)
					return pt, url, false
				end
			else
				local advagain = false
				if pt:contains([[<a href="adventure.php?snarfblat=]]..zoneid..[[">Adventure Again]]) then
					advagain = true
				else
					print("non-fight non-choice unhandled url", url)
				end
	--~ 			print("return4 p u a", pt:len(), url, advagain)
				return pt, url, advagain
			end
		end

		function autoadventure(zoneid, macro, noncombatchoices, specialnoncombatfunction)
			set_fight_state("zone", zoneid)
			set_ascension_state("adventure.lastzone", zoneid)
			if not noncombatchoices then noncombatchoices = {} end
			local pt, url = post_page("/adventure.php", { snarfblat = tostring(zoneid) })
			return handle_adventure_result(pt, url, zoneid, macro, noncombatchoices, specialnoncombatfunction)
		end

		function get_skills()
			local cs = get_page("/charsheet.php")
			local skills_text = cs:match("<p>Skills:</b>.-(<a onClick.-)</td>")
			local skills = {}
			for s in skills_text:gmatch("<a onClick.->(.-)</a>") do
				skills[s] = true
			end
			return skills
		end
	end

	local s_f_env = {}
	setmetatable(s_f_env, { __index = _G, __newindex = _G })
	setfenv(setup_functions_raw, s_f_env)
	setup_functions_raw()
	setmetatable(s_f_env, { __index = _G,
	__newindex = function(t, k, v)
		print("warning: s_f_env __newindex", t, k, v)
--~ 		_G[k] = v
		rawset(t, k, v)
	end })
end

function setup_variables()
	if path == "/charpane.php" then
		familiar_picture, familiar_weight = text:match([[<a target=mainpane href="familiar.php" class="familiarpick"><img src="http://images.kingdomofloathing.com/itemimages/([^\"]-).gif" width=30 height=30 border=0></a><br>([0-9]+) lb]])
		if not familiar_picture then
			familiar_picture, familiar_weight = text:match([[<a target=mainpane href="familiar.php" class="familiarpick"><img src="http://images.kingdomofloathing.com/itemimages/([^\"]-).gif" width=30 height=30 border=0>.- <b>([0-9]+)</b> pound ]]) -- TODO-future: redo without .-?
		end
		familiar_weight = tonumber(familiar_weight)
		if text:match([[<td align=right>Meat:.-</table>]]) then
			compact_charpane = true
		elseif text:match([[src="http://images.kingdomofloathing.com/itemimages/hourglass.gif".-</table>]]) then
			normal_charpane = true
		end
	end

	for x in text:gmatch("var onturn = ([0-9]+);") do
		combat_round = tonumber(x)
	end

	monster_name = text:match("<span id='monname'>(.-)</span>")

	if path == "/fight.php" then
		local z = get_fight_state("zone")
		if z ~= "" then
			adventure_zone = tonumber(z)
		end
	end

	if monster_name and query:match("ireallymeanit") then
		new_fight_started = true
	end

	adventure_title = text:match([[<tr><td style="color: white;" align=center bgcolor=blue><b>([^<]*)</b></td></tr>]])
	choice_adventure_number = tonumber(text:match([[<input type=hidden name=whichchoice value=([0-9]+)>]]))
	adventure_result = text:match([[<td style="color: white;" align=center bgcolor=blue><b>Adventure Results:</b></td></tr><tr><td style="padding: 5px; border: 1px solid blue;"><center><table><tr><td><center><b>([^<]*)</b>]])

	fight_is_over = nil
	if path == "/fight.php" then
		if text:contains("state['fightover'] = true;") then -- TODO: HACK! this only works with combat bar enabled!!
			fight_is_over = true
		else
			fight_is_over = false
		end
	end
end

function run_functions(path, pagetext, run)
	pagetext = run(path, pagetext)

	if path == "/fight.php" then
		pagetext = pagetext:gsub([[(<td[^>]-><img src="http://images.kingdomofloathing.com/itemimages/)([^"]+.gif)(" width=30 height=30 alt="[^"]+" title=")([^"]+)("></td><td[^>]->)(.-)(</td></tr>)]], function (pre, itemimage, mid, title, td, msg, post)
	--~ 		print("item use", msg)
			item_image = itemimage
			item_name = title
			msg = run("used combat item", msg)
			item_image = nil
			item_name = nil
			return pre .. itemimage .. mid .. title .. td .. msg .. post
		end)

		pagetext = pagetext:gsub([[<center><table class="item" style="float: none" rel="[^"]*"><tr><td><img src="http://images.kingdomofloathing.com/itemimages/[^"]+.gif" alt="[^"]*" title="[^"]*" class=hand onClick='descitem%([0-9]+%)'></td><td valign=center class=effect>You acquire .-</td></tr></table></center>]], function (droptext)
--~ 			print("item capture", pre, rel, mid, dropinfo, post)
			item_image = droptext:match([[src="http://images.kingdomofloathing.com/itemimages/([^"]+).gif"]])
			item_name = droptext:match([[title="([^"]*)"]])
			msg = droptext
			if item_name then
				msg = run("item drop: " .. item_name, msg)
			end
			msg = run("item drop", msg)
	-- 		print("item capture", pre, rel, mid, dropinfo, post)
	--~ 		local msg = pre .. rel .. mid .. "<span style=\"color: darkgreen\">" .. dropinfo .. "</span>" .. post
	--~ 		if string.match(rel, "u=u") then
	--~ 			msg = pre .. rel .. mid .. "<span style=\"color: darkgreen\">" .. dropinfo .. "</span> [use]" .. post
	--~ 			http://localhost:18481/inv_use.php?pwd=xxx&which=3&whichitem=3236
	--~ 		elseif string.match(rel, "u=q") then
	--~ 			msg = pre .. rel .. mid .. "<span style=\"color: darkgreen\">" .. dropinfo .. "</span> [equip]" .. post
	--~ 			http://localhost:18481/inv_equip.php?pwd=xxx&which=2&action=equip&whichitem=2813
	--~ 		end
			return msg
		end)

		pagetext = pagetext:gsub([[(<!%-%-familiarmessage%-%-><center><table>.-</table></center>)]], function (msg)
			familiarmessage_picture = msg:match([[<!%-%-familiarmessage%-%-><center><table><tr><td align=center valign=center><img src="http://images.kingdomofloathing.com/itemimages/([^"]+).gif" width=30 height=30></td>]])
			if familiarmessage_picture then
				msg = run("familiar message: " .. familiarmessage_picture, msg)
			end
			msg = run("familiar message", msg)
			return msg
		end)
	end

	pagetext = run("all pages", pagetext)

	return pagetext
end
